"""
Name        : multiListbox.py
Author      : Josef Laimer
Version     :
Copyright   : Copyright (C) 2011 Josef Laimer, Peter Lackner, University of Salzburg
Description : Sortable multilistbox based on an implemention from Brent Burley
              http://code.activestate.com/recipes/52266/
Notes       : 
"""



from Tkinter import Frame, Button, Listbox, Label, Scrollbar
from Tkconstants import *  # @UnusedWildImport
from types import *  # @UnusedWildImport
import traceback



class MultiListbox(Frame):
    """
    @param master: master frame
    @type master: tkinter container
    @param multiselection: True if a selection over more than one row should be allowed.
    @type multiselection: Boolean
    @param lists: Tuple which contains column informations (see below *)
    @type lists: tuple of tuple
    @param callback: callback function which should be called after a selection change
    @type callback: function
    @param callbackDeSel: callback function which should be called after a selection is deselected
    @type callback: function
    
    *column description: 
    tuple: (column name, column width, type, justification)
    column name: string
    column width: integer
    type: IntType, LongType, FloatType, BooleanType, or StringType
    justification
    
    edit for Maestro/Pirates: added a second functionpointer which is executed if a element is deselected
    
    """
    def __init__(self, master, multiselection, lists, callback = None,callbackDeSel = None):
        Frame.__init__(self, master)
        self.lists = []     #one list per column
        self.colmapping={}  #mapping between header buttons and columns (lists)
        self.colItemTypes = {} #mapping between columns and data type
        self.justifyList = [] #mapping between columns and justify typ
        self.maxLength = [] #column width ########TODO
#        self.origData = None
        self.multiSelection = multiselection #determines if the list allows more than one selected item 
                                             
        self.lastSelection = None
        
        self.callback = callback
        self.callbackDeSel = callbackDeSel
                
        for l,w,t,j in lists:
            frame = Frame(self); frame.pack(side=LEFT, expand=YES, fill=BOTH)
            b = Button(frame, text=l, borderwidth=1, relief=RAISED)
            b.pack(fill=X)
            b.bind('<Button-1>', self._sort)
            self.colmapping[b]=(len(self.lists),1)
            
            lb = Listbox(frame, width=w, borderwidth=0, selectborderwidth=0,
                         relief=FLAT, exportselection=FALSE)
            lb.pack(expand=YES, fill=BOTH)
            self.colItemTypes[lb] = t
            self.justifyList.append(j)
            self.maxLength.append(w + 10)  ########TODO
            self.lists.append(lb)
            lb.bind('<B1-Motion>', lambda e, s=self: s._select(e.y))
            lb.bind('<Button-1>', lambda e, s=self: s._select(e.y))
            lb.bind('<Leave>', lambda e: 'break')
            lb.bind('<B2-Motion>', lambda e, s=self: s._b2motion(e.x, e.y))
            lb.bind('<Button-2>', lambda e, s=self: s._button2(e.x, e.y))
            lb.bind('<Button-4>', lambda e, s=self: s._scroll(SCROLL, -1, UNITS))
            lb.bind('<Button-5>', lambda e, s=self: s._scroll(SCROLL, 1, UNITS))
            lb.bind('<MouseWheel>', lambda e, s=self: s._on_mousewheel(e))
        frame = Frame(self); frame.pack(side=LEFT, fill=Y)
        Label(frame, borderwidth=1, relief=RAISED).pack(fill=X)
        sb = Scrollbar(frame, orient=VERTICAL, command=self._scroll)
        sb.pack(expand=YES, fill=Y)
        self.lists[0]['yscrollcommand']=sb.set
   
    def setcallback (self, callback):        
        self.callback = callback
       
    def justify(self):        
        #get data from the lists
        tableData = self.get(0, END)
        #get numb. of list entries (rows)
        rowcount = len(tableData)
        
        #helper table 
        curTable = []
        
        #justify the list entries
        for row in range(rowcount):            
            curTable.append([])
            for col in range(len(tableData[row])):
                if tableData[row][col] != None:
                    if self.justifyList[col] == RIGHT:
                        colString = tableData[row][col]
                        colString = colString.rjust(self.maxLength[col], ' ')                    
                        curTable[row].append(colString)
                    elif self.justifyList[col] == CENTER:
                        colString = tableData[row][col]
                        colString = colString.center(self.maxLength[col], ' ')                    
                        curTable[row].append(colString)
                    else: 
                        curTable[row].append(tableData[row][col]) 
                else: 
                    curTable[row].append(tableData[row][col])
                    
        #clear widget
        self.delete(0, END)
        
        #refill widgets
        for row in range(rowcount):
            self.insert(END, curTable[row])
        
        
    def _sort(self, e):
        # get the listbox to sort by (mapped by the header button)
        b=e.widget
        col, direction = self.colmapping[b]

        isNumberCol = False
        if self.colItemTypes[self.lists[col]] == IntType or self.colItemTypes[self.lists[col]] == FloatType:
            isNumberCol = True           

        # get the entire table data into mem
        if isNumberCol:
            tableData = self.getOrigDatatype(0,END)
        else:
            tableData = self.get(0, END)
        
#        if self.origData == None:
#            import copy
#            self.origData = copy.deepcopy(tableData)


        rowcount = len(tableData)

        #remove old sort indicators if it exists
        for btn in self.colmapping.keys():
            lab = btn.cget('text')
            if lab[0]=='[': btn.config(text=lab[4:])

        btnLabel = b.cget('text')
        
        
        #sort data based on direction
        if direction==0:
            def colsort(x, y, mycol=0, direction=1):
                return direction*cmp(x[mycol], y[mycol])
            
            tableData.sort(colsort)
            
            #tableData = self.origData
        else:
            if direction==1: b.config(text='[+] ' + btnLabel)
            else: b.config(text='[-] ' + btnLabel)
            # sort by col
            def colsort(x, y, mycol=col, direction=direction):
                return direction*cmp(x[mycol], y[mycol])

            tableData.sort(colsort)

        #clear widget
        self.delete(0,END)

        # refill widget
        for row in range(rowcount):
            self.insert(END, tableData[row])

        # toggle direction flag 
        if(direction==1): direction=-1
        else: direction += 1
        self.colmapping[b] = (col, direction)
        
        if isNumberCol:
            self.justify()
        
  
    def _select(self, y):
        row = self.lists[0].nearest(y)
        if self.selection_includes(row):
            self.selection_clear(row, row)
        else:
            if not self.multiSelection: 
                self.selection_clear(0, END)
            self.selection_set(row)
        return 'break'

    def _button2(self, x, y):
        for l in self.lists: l.scan_mark(x, y)
        return 'break'

    def _b2motion(self, x, y):
        for l in self.lists: l.scan_dragto(x, y)
        return 'break'

    def _scroll(self, *args):
        for l in self.lists:
            apply(l.yview, args)
        return 'break'
    #------------------------------------------------------------
    def _on_mousewheel(self, event):
        for l in self.lists:
            l.yview_scroll(-1*(event.delta/120), "units")
        return "break"
    #------------------------------------------------------------
    
    def curselection(self):
        return self.lists[0].curselection()

    def delete(self, first, last=None):
        tmpList = self.get(first, last)
        tmpList     
        for l in self.lists:
            l.delete(first, last)
 
            
    def deleteAll(self):
        if self.lists[0].size() > 0: 
            self.delete(0, END)
        
    def deleteSelection(self):
        curSel = self.lists[0].curselection()
        for sel in curSel: 
            self.delete(sel, sel)
        

    def get(self, first, last=None):
        result = []
        for l in self.lists:                        
            result.append(l.get(first,last))
        if last: return apply(map, [None] + result)
        return result
    
    def getOrigDatatype (self, first, last=None):
        result = []
        for l in self.lists:
            data = l.get(first, last)
            dataTmp = []
            datatype = self.colItemTypes[l]
            if datatype == IntType:
                for i in range(len(data)):
                    intVar = data[i].strip()
                    if intVar == '':
                        dataTmp.append(None)
                    else:
                        intVar = int(eval(intVar))
                        dataTmp.append(intVar)
            elif datatype == FloatType:
                for i in range(len(data)):
                    flo = data[i].strip()
                    if flo == '':
                        dataTmp.append(None)
                    else:
                        flo = float(eval(flo))
                        dataTmp.append(flo)
            else:
                dataTmp = data         
                        
            result.append(dataTmp)
        if last: return apply(map, [None] + result)
        return result
        
    def index(self, index):
        self.lists[0].index(index)

    def insert(self, index, *elements):
        for e in elements:
            i = 0            
            for l in self.lists:
                if e[i] == None:
                    l.insert(index, '')
                else:
                    l.insert(index, str(e[i]))
                i = i + 1

    def size(self):
        return self.lists[0].size()

    def see(self, index):
        for l in self.lists:
            l.see(index)

    def selection_anchor(self, index):
        for l in self.lists:
            l.selection_anchor(index)

    
    def selection_clear(self, first, last=None):
        for l in self.lists:
            l.selection_clear(first, last)
            self.lastDeselection = first
        if self.callbackDeSel != None:
            #-------------------------------------------------------
            try:
                self.callbackDeSel()
            except Exception as e:
                print e
                traceback.print_exc(file=sys.stdout)

    def selection_includes(self, index):
        return self.lists[0].selection_includes(index)

    def selection_set(self, first, last=None):
        for l in self.lists:
            l.selection_set(first, last)
            self.lastSelection = first
        if self.callback != None:
            #-------------------------------------------------------
            try:
                self.callback()
            except Exception as e:
                print e
                traceback.print_exc(file=sys.stdout)
    
    def selectLastEntry(self):
        if not self.multiSelection: 
            self.selection_clear(0, END)
        self.selection_set(END)
        for list_ in self.lists:
            list_.see(int(list_.curselection()[-1]))
        return 'break'
        
    #returns the values of the first column (ids) 
    def selected_datasets(self):
        curSel = self.lists[0].curselection() 
        selectedIds = []
        for sel in curSel:
            selectedIds.append(self.lists[0].get(sel)) 
            
        return selectedIds
    
    def editRowBySelectionIndex (self, selIndex, colIndex, text):
        rowIndex = self.lists[colIndex].curselection()[selIndex]
        self.lists[colIndex].delete(rowIndex)
        self.lists[colIndex].insert(rowIndex, text)    
        self.lists[colIndex].selection_set(rowIndex)
    
    #returns the last selected entry (id)
    def getLastSelection (self):
        if self.lists[0].size() > self.lastSelection:
            return self.lists[0].get(self.lastSelection)
        #returns the last selected entry (id)
    def getLastSelectionSec (self):
        if self.lists[1].size() > self.lastSelection:
            return self.lists[1].get(self.lastSelection)
        
    #returns the last selected entry (id)
    def getLastDeSelection (self):
        """
        returns the last deselected item
        """
        if self.lists[0].size() > self.lastDeselection:
            return self.lists[0].get(self.lastDeselection)
